#####################################################################
# -*- coding: iso-8859-1 -*-                                        #
#                                                                   #
# Frets on Fire                                                     #
# Copyright (C) 2006 Sami Kystil                                  #
#                                                                   #
# This program is free software; you can redistribute it and/or     #
# modify it under the terms of the GNU General Public License       #
# as published by the Free Software Foundation; either version 2    #
# of the License, or (at your option) any later version.            #
#                                                                   #
# This program is distributed in the hope that it will be useful,   #
# but WITHOUT ANY WARRANTY; without even the implied warranty of    #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     #
# GNU General Public License for more details.                      #
#                                                                   #
# You should have received a copy of the GNU General Public License #
# along with this program; if not, write to the Free Software       #
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,        #
# MA  02110-1301, USA.                                              #
#####################################################################

from Scene import SceneServer, SceneClient
import os
import Player
import Dialogs
import Song
import Config
from Language import _

# save chosen song into config file
Config.define("game", "selected_song",     str, "")

class SongChoosingScene:
  pass

class SongChoosingSceneServer(SongChoosingScene, SceneServer):
  pass

class SongChoosingSceneClient(SongChoosingScene, SceneClient):
  def createClient(self, songName = None):
    self.wizardStarted = False
    self.songName      = songName

  def freeResources(self):
    self.songs = None
    self.cassette = None
    self.folder = None
    self.label = None
    self.song = None
    self.background = None
    
  def run(self, ticks):
    SceneClient.run(self, ticks)

    if not self.wizardStarted:
      self.wizardStarted = True
      library = Config.get("game", "selected_library")
      
      if not library or library == "None" or not os.path.exists(self.engine.resource.fileName(library)):
        library = "songs"
        Config.set("game", "selected_library", library)
      
      files = os.listdir(self.engine.resource.fileName(library))
      
      if not self.songName:
        while True:
          self.songName = Dialogs.chooseSong(self.engine, selectedSong = Config.get("game", "selected_song"))

          if not self.songName:
            newlibrary = os.path.dirname(library)
            if not newlibrary:    
              self.session.world.finishGame()
              return
            else:
              library = newlibrary
              Config.set("game", "selected_library", library)
              continue

          Config.set("game", "selected_song", self.songName)
          info = Song.loadSongInfo(self.engine, self.songName)
          if info.artist == "=FOLDER=":
            library = os.path.join(library, self.songName)
            Config.set("game", "selected_library", library)
            continue

          if len(info.difficulties) > 1:
            d = Dialogs.chooseItem(self.engine, info.difficulties,
                                 _("Choose a difficulty:"), selected = self.player.difficulty)
          else:
            d = info.difficulties[0]
          if d:
            #Dialogs.showLoadingScreen(self.engine, lambda: info.parts, text = _("Checking The Band..."))
            self.player.difficulty = d
            if len(info.parts) > 1:
              p = Dialogs.chooseItem(self.engine, info.parts,
                                 _("Choose a part:"), selected = self.player.part)
            else:
              p = info.parts[0]
              
            if p:
              self.player.part = p
              break
      else:
        info = Song.loadSongInfo(self.engine, self.songName)

      # Make sure the difficulty we chose is available
      if not self.player.difficulty in info.difficulties:
        self.player.difficulty = info.difficulties[0]
      if not self.player.part in info.parts:
        self.player.part = info.parts[0]
      
      self.session.world.deleteScene(self)
      self.freeResources()
      self.session.world.createScene("GuitarScene", songName = self.songName)
